package ejb;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Set;

import javax.ejb.Stateless;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;

import jpa.ActividadJPA;
import jpa.ActividadUpdateJPA;
import jpa.FaseJPA;
import jpa.FaseUpdateJPA;
import jpa.ListaJPA;
import jpa.ParteJPA;
import jpa.ParteUpdateJPA;
import jpa.RequisitoJPA;
import jpa.RequisitoUpdateJPA;
import jpa.TipoActividadJPA;
import jpa.TipoFaseJPA;
import jpa.TipoListaJPA;
import jpa.TipoRequisitoJPA;
import jpa.UsuarioJPA;
import rest.RespuestaRest;

@Stateless
public class DTFacadeBeanRemote implements DTFacadeRemote {

	@PersistenceContext(unitName = "primary")
	private EntityManager entman;

	/**
	 * Mètode que retorna un usuari transportista que es correspongui amb el login i el password d'entrada.
	 */
	@Override
	public RespuestaRest<UsuarioJPA> login(String usr, String pass) {
		
		RespuestaRest<UsuarioJPA> respuesta = new RespuestaRest<>();
		UsuarioJPA usuario = null;
				
		try {
			usuario = (UsuarioJPA) entman.createQuery("from UsuarioJPA u where u.rol = 'transportista' and u.login = :login and u.password = :password").setParameter("login", usr).setParameter("password", pass).getSingleResult();
			usuario.setToken("token a l'atzar");
			respuesta.setCodigo("200");
			respuesta.setMensaje("Usuario autenticado correctamente");
		} catch (Exception e) {
			respuesta.setCodigo("403");
			respuesta.setMensaje("Autenciación incorrecta");
		}
		respuesta.setDatos(usuario);
		
		return respuesta;
	}
	
	/**
	 * Mètode que desautentica un usuari transportista de l'aplicació mòbil.
	 */
	// TODO Només hi ha la estructura perquè falta implementar un control mitjançant token.
	@Override
	public RespuestaRest<String> logout(int idUsuario) {
		
		RespuestaRest<String> respuesta = new RespuestaRest<>();
		UsuarioJPA usuario = null;
		try {
			usuario = entman.find(UsuarioJPA.class, idUsuario);
			usuario.setToken("");
			respuesta.setCodigo("200");
			respuesta.setMensaje("Usuario desconectado correctamente");
		} catch (Exception e) {
			respuesta.setCodigo("403");
			respuesta.setMensaje("Desconexión incorrecta");
		}
		
		return respuesta;
	}

	/**
	 * Mètode que retorna l'últim (per ordre descendent de id) comunicat de treball OBERT de l'usuari amb l'id d'entrada.
	 */
	public RespuestaRest<ParteJPA> ultimoParte(int id) {
		
		RespuestaRest<ParteJPA> respuesta = new RespuestaRest<>();
		respuesta.setTipoDatos("parte");

		UsuarioJPA usuario = entman.find(UsuarioJPA.class, id);
		if (usuario != null) {
			try {
				ParteJPA parte = (ParteJPA) entman
						.createQuery(
								"from ParteJPA p LEFT OUTER JOIN FETCH p.actividades a LEFT OUTER JOIN FETCH a.listas l LEFT OUTER JOIN FETCH l.fases f LEFT OUTER JOIN FETCH f.requisitos r where p.conductor.id = :id and p.abierto = 1 order by p.id desc")
						.setParameter("id", usuario.getId()).getResultList().get(0);
				respuesta.setCodigo("200");
				respuesta.setDatos(parte);
			} catch (Exception e) {
				respuesta.setCodigo("500");
				respuesta.setMensaje("Error al recuperar el último parte de trabajo. Inténtelo de nuevo más tarde.");
			}
		}

		return respuesta;
	}

	/**
	 * Mètode que genera i retorna un nou comunicat de treball per l'usuari amb l'id d'entrada. També registra la data i la hora de l'acció.
	 * Si l'usuari tenia l'últim comunicat obert, el tanca junt amb les seves activitats i fases.
	 */
	public RespuestaRest<ParteJPA> nuevoParte(int idUsuario, String coordenadas) {
		
		RespuestaRest<ParteJPA> respuesta = new RespuestaRest<>();
		respuesta.setTipoDatos("parte");

		UsuarioJPA usuario = entman.find(UsuarioJPA.class, idUsuario);
		if(usuario != null){
			ParteJPA parte = ultimoParte(idUsuario).getDatos();
			if (parte != null) {
				RespuestaRest<String> resp = cerrarParte(parte.getId(), coordenadas);
				if (!resp.getCodigo().equals("200")) {
					respuesta.setCodigo("500");
					respuesta.setMensaje("Error al cerrar el último parte de trabajo. Inténtelo de nuevo más tarde.");
					return respuesta;
				}
			}
			parte = new ParteJPA();
			parte.setConductor(usuario);
			parte.setInicio(new SimpleDateFormat("dd/MM/yyyy HH:mm").format(new Date()));
			parte.setCoordenadasInicio(coordenadas);
			entman.persist(parte);
			usuario.getPartes().add(parte);
			respuesta.setCodigo("200");
			respuesta.setMensaje("Nuevo parte creado correctamente. Proceda a su cumplimentación.");
			parte = ultimoParte(idUsuario).getDatos();
			respuesta.setDatos(parte);
		} else {
			respuesta.setCodigo("500");
			respuesta.setMensaje("Error intentar abrir un nuevo parte de trabajo. Inténtelo de nuevo más tarde.");
		}
		
		return respuesta;
	}

	/**
	 * Mètode que actualitza el comunicat de treball amb les dades del JSON d'entrada.
	 */
	@Override
	public RespuestaRest<ParteJPA> actualizarParte(ParteUpdateJPA parteJson) {
		
		RespuestaRest<ParteJPA> respuesta = new RespuestaRest<>();
		respuesta.setTipoDatos("parte");
		
		if (parteJson != null){
			entman.merge(parteJson);
			ParteJPA parte = verParte(parteJson.getId());
			respuesta.setCodigo("200");
			respuesta.setMensaje("Parte actualizado correctamente");
			respuesta.setDatos(parte);
		} else {
			respuesta.setCodigo("500");
			respuesta.setMensaje("Error al actualizar el parte de trabajo. Inténtelo de nuevo más tarde.");
		}
		
		
		return respuesta;
	}

	/**
	 * Mètode que canvia a 'tancat' l'estat del comunicat amb l'id d'entrada, així com de les seves activitats i fases. També registra la data, hora i coordenades de l'acció.
	 */
	@Override
	public RespuestaRest<String> cerrarParte(int id, String coordenadas) {
		
		RespuestaRest<String> respuesta = new RespuestaRest<>();
		respuesta.setTipoDatos("parte");
		
		ParteJPA parte = entman.find(ParteJPA.class, id);
		if(parte != null){
			parte.setAbierto(false);
			parte.setFin(new SimpleDateFormat("dd/MM/yyyy HH:mm").format(new Date()));
			parte.setCoordenadasFin(coordenadas);
			cerrarActividades(parte.getActividades(), coordenadas);
			respuesta.setCodigo("200");
			respuesta.setMensaje("Parte cerrado correctamente");
		} else {
			respuesta.setCodigo("500");
			respuesta.setMensaje("Error al cerrar el parte de trabajo. Inténtelo de nuevo más tarde.");
		}
		
		return respuesta;
	}

	/**
	 * Mètode que genera i retorna una nova activitat pel comunicat amb l'id d'entrada. També registra la data, hora i coordenades de l'acció.
	 * Si el comunicat tenia alguna activitat oberta, la tanca junt amb les seves fases.
	 */
	@Override
	public RespuestaRest<ParteJPA> nuevaActividad(int idParte, String coordenadas, int tipo) {
		
		RespuestaRest<ParteJPA> respuesta = new RespuestaRest<>();
		respuesta.setTipoDatos("parte");

		ParteJPA parte = entman.find(ParteJPA.class, idParte);
		if(parte != null){
			TipoActividadJPA tipoActividad = entman.find(TipoActividadJPA.class, tipo);
			if(tipoActividad != null){
				cerrarActividades(parte.getActividades(), coordenadas);
				ActividadJPA actividad = new ActividadJPA();
				actividad.setTipo(tipoActividad);
				actividad.setParte(parte);
				actividad.setInicio(new SimpleDateFormat("dd/MM/yyyy HH:mm").format(new Date()));
				actividad.setCoordenadasInicio(coordenadas);
				entman.persist(actividad);
				parte.getActividades().add(actividad);
				if (tipo == 2) {
					nuevaLista(actividad, 1);
				}
				respuesta.setCodigo("200");
				respuesta.setMensaje("Nueva actividad creada correctamente. Proceda a su cumplimentación.");
				parte = verParte(idParte);
				respuesta.setDatos(parte);
				System.out.println(parte);
			} else {
				respuesta.setCodigo("500");
				respuesta.setMensaje("Error al localizar el tipo de actividad. Inténtelo de nuevo más tarde.");
			}
		} else {
			respuesta.setCodigo("500");
			respuesta.setMensaje("Error al localizar el parte de trabajo. Inténtelo de nuevo más tarde.");
		}
		
		return respuesta;
	}

	/**
	 * Mètode que actualitza l'activitat del comunicat de treball amb les dades del JSON d'entrada.
	 */
	@Override
	public RespuestaRest<ParteJPA> actualizarActividad(int idParte, ActividadUpdateJPA actividadJson) {
	
		RespuestaRest<ParteJPA> respuesta = new RespuestaRest<>();
		respuesta.setTipoDatos("parte");
		
		if (actividadJson != null){
			entman.merge(actividadJson);
			ParteJPA parte = verParte(idParte);
			respuesta.setCodigo("200");
			respuesta.setMensaje("Actividad actualizada correctamente");
			respuesta.setDatos(parte);
		} else {
			respuesta.setCodigo("500");
			respuesta.setMensaje("Error al actualizar la actividad. Inténtelo de nuevo más tarde.");
		}
		
		return respuesta;
	}

	/**
	 * Mètode que canvia a 'tancat' de l'activitat amb l'id d'entrada, així com de les seves fases. També registra la data, hora i coordenades de l'acció.
	 */
	@Override
	public RespuestaRest<ParteJPA> cerrarActividad(int idActividad, String coordenadas) {
		
		RespuestaRest<ParteJPA> respuesta = new RespuestaRest<>();
		respuesta.setTipoDatos("parte");
		
		ActividadJPA actividad = entman.find(ActividadJPA.class, idActividad);
		if(actividad != null){
			actividad.setAbierta(false);
			actividad.setFin(new SimpleDateFormat("dd/MM/yyyy HH:mm").format(new Date()));
			actividad.setCoordenadasFin(coordenadas);
			for (ListaJPA lista : actividad.getListas()) {
				cerrarFases(lista.getFases());
			}
			respuesta.setCodigo("200");
			respuesta.setMensaje("Actividad cerrada correctamente");
			respuesta.setDatos(verParte(actividad.getParte().getId()));
		} else {
			respuesta.setCodigo("500");
			respuesta.setMensaje("Error al cerrar la actividad. Inténtelo de nuevo más tarde.");
		}
		
		return respuesta;
	}

	/**
	 * Mètode que actualitza la fase del comunicat de treball amb les dades del JSON d'entrada.
	 */
	@Override
	public RespuestaRest<ListaJPA> actualizarFase(int idParte, FaseUpdateJPA faseJson) {
		
		RespuestaRest<ListaJPA> respuesta = new RespuestaRest<>();
		respuesta.setTipoDatos("lista");
		
		if (faseJson != null){
			Set<RequisitoUpdateJPA> requisitos = faseJson.getRequisitos();
			for(RequisitoUpdateJPA requisito : requisitos){
				entman.merge(requisito);
			}
			FaseJPA fase = entman.find(FaseJPA.class, faseJson.getId());
			ListaJPA lista = (ListaJPA) entman.createQuery("from ListaJPA l LEFT OUTER JOIN FETCH l.fases f LEFT OUTER JOIN FETCH f.requisitos r where l.id = :id order by f.id, r.id")
					.setParameter("id", fase.getLista().getId()).getSingleResult();
			respuesta.setCodigo("200");
			respuesta.setMensaje("Fase actualizada correctamente");
			respuesta.setDatos(lista);
		} else {
			respuesta.setCodigo("500");
			respuesta.setMensaje("Error al actualizar la fase. Inténtelo de nuevo más tarde.");
		}
		
		return respuesta;
	}

	/**
	 * Mètode que canvia a 'tancat' l'estat de la fase amb l'id d'entrada.
	 */
	@Override
	public RespuestaRest<ListaJPA> siguienteFase(int idParte, int idFase) {
		
		RespuestaRest<ListaJPA> respuesta = new RespuestaRest<>();
		respuesta.setTipoDatos("lista");
		
		FaseJPA fase = entman.find(FaseJPA.class, idFase);
		if(fase != null){
			fase.setAbierta(false);
			ListaJPA lista = (ListaJPA) entman.createQuery("from ListaJPA l LEFT OUTER JOIN FETCH l.fases f LEFT OUTER JOIN FETCH f.requisitos r where l.id = :id order by f.id, r.id")
					.setParameter("id", fase.getLista().getId()).getSingleResult();
			respuesta.setCodigo("200");
			respuesta.setMensaje("Siguiente fase activada correctamente");
			respuesta.setDatos(lista);
		} else {
			respuesta.setCodigo("500");
			respuesta.setMensaje("Error al activar la siguiente fase. Inténtelo de nuevo más tarde.");
		}
		
		
		
		return respuesta;
	}
	
	/**
	 * Mètode que retorna un objecte ParteJPA amb totes les seves entitats relacionades per l'id d'entrada.
	 */
	private ParteJPA verParte(int id){
		
		ParteJPA parte = (ParteJPA) entman.createQuery("from ParteJPA p LEFT OUTER JOIN FETCH p.actividades a LEFT OUTER JOIN FETCH a.listas l LEFT OUTER JOIN FETCH l.fases f LEFT OUTER JOIN FETCH f.requisitos r where p.id = :id order by a.id, l.id, f.id, r.id")
						.setParameter("id", id).getSingleResult();
		return parte;
		
	}

	/**
	 * Mètode que crea una llista a partir de una plantilla. De moment es de tipus fixe.
	 * @param actividad Activitat per la que es crea la llista
	 * @param tipo		Tipus de llista a crear
	 */
	// TODO Falta crear les pantalles i funcionalitats necessaries per gestionar les plantilles de llistes, fases i requisits. 
	private void nuevaLista(ActividadJPA actividad, int tipo) {
	
		TipoListaJPA tl = entman.find(TipoListaJPA.class, tipo);
		if (tl != null) {
			ListaJPA lista = new ListaJPA();
			lista.setNombre(tl.getNombre());
			lista.setMaxFases(tl.getMaxFases());
			lista.setTipoLista(tl.getId());
			entman.persist(lista);
			actividad.getListas().add(lista);
	
			Set<TipoFaseJPA> stf = tl.getTipos_fase();
			for (TipoFaseJPA tf : stf) {
				FaseJPA fase = new FaseJPA();
				fase.setNombre(tf.getNombre());
				fase.setOrden(tf.getOrden());
				entman.persist(fase);
				lista.getFases().add(fase);
	
				Set<TipoRequisitoJPA> str = tf.getTipos_requisito();
				for (TipoRequisitoJPA tr : str) {
					RequisitoJPA requisito = new RequisitoJPA();
					requisito.setNombre(tr.getNombre());
					requisito.setAyuda(tr.getAyuda());
					entman.persist(requisito);
					fase.getRequisitos().add(requisito);
				}
			}
		}
	}
	
	/**
	 * Mètode que canvia a 'tancat' l'estat de una llista d'activitats i de les seves fases. També registra la data, hora i coordenades de l'acció.
	 * @param actividades
	 * @param coordenadas
	 */
	private void cerrarActividades(Set<ActividadJPA> actividades, String coordenadas){
		for(ActividadJPA actividad : actividades){
			if (actividad.getAbierta()) {
				actividad.setAbierta(false);
				actividad.setFin(new SimpleDateFormat("dd/MM/yyyy HH:mm").format(new Date()));
				actividad.setCoordenadasFin(coordenadas);
				for (ListaJPA lista : actividad.getListas()) {
					cerrarFases(lista.getFases());
				}
			}
			
		}
	}
	
	/**
	 * Mètode que canvia a 'tancat' l'estat de una llista de fases.
	 * @param fases
	 */
	private void cerrarFases(Set<FaseJPA> fases){
		for (FaseJPA fase : fases) {
			if (fase.getAbierta()) {
				fase.setAbierta(false);
			}
		}
	}
	
}
